/*
 * example.c -- demo for the libqsearch API.
 *              see http://www.cartel-securite.net/pbiondi/libqsearch.html
 *              for more informations
 *
 * Copyright (C) 2002  Philippe Biondi <biondi@cartel-securite.fr>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <dlfcn.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include "qsearch.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#define SCANSIZE 8192

typedef struct pats {
	char *ptrn;
	char *descr;
	int type;
} pats;

pats pat[] = { 
	{"123123412345","123123412345", QSEARCH_CS},
	{"4123456",	"4123456",	QSEARCH_CS},
	{"123456",	"123456",	QSEARCH_CS},
	{"45123",	"45123",	QSEARCH_CS},
	{"efghih", 	"efghih",	QSEARCH_CS},
	{"efghi", 	"efghi -CI",	QSEARCH_CI},
	{"ab\ncd",	"ab<LF>cd",	QSEARCH_CS},
	{"ab\ncd\nef",	"ab<LF>cd<LF>ef -CI", QSEARCH_CI},
	{"c*n*rd",	"c*n*rd",	QSEARCH_JOKER_CS},
	{"*e\nt*rd",	"*e<LF>t*rd",	QSEARCH_JOKER_CS},
        {"t*paze",	"t*paze -CI",	QSEARCH_JOKER_CI},
	{NULL,NULL,0}
};
		

static int cb(void *cb_data, void *ptrn_data, size_t match_offset)
{
	printf("Found [%s] at pos %i in buffer #%i\n",(char *)ptrn_data, match_offset, *(int *)cb_data);
	return 0;
}

static void usage()
{
	char **plugin_list, **p;
    	fprintf(stderr,"Error: Usage: example -p <plugin> [-a <algo>] [-f <pattern file> [-t <type>]] \n                      [-j <joker>]\n");

	plugin_list=qsearch_get_plugin_list();
	if (!plugin_list) 
		goto the_end;
	
	fprintf(stderr,"Available plugins :\n");
	for (p=plugin_list; *p; p++) {
		fprintf(stderr," + %s\n",*p);
	}
	qsearch_free_plugin_list(plugin_list);

the_end:	
	exit(1);
}

int main(int argc, char **argv) 
{
	qsearch_plugin_t *plugin;
	qsearch_algo_t *algo, **list;
	qsearch_obj_t *obj;
	qsearch_options_t options;
	qsearch_state_t *state_in, *state_out, *state_io;

	int ret;
	size_t bufofs,size;
	void *buf;
	pats *idxp;
        struct timeval before,after;
	double duration;
	int patnb=0,c;
	char *patternfile=NULL, *algoname=NULL, *pluginname=NULL;
	FILE *f=NULL;
	int type=QSEARCH_CS;
	int count;
	char joker='*';

	/* Parsing arguments line */
	while ((c = getopt(argc,argv, "p:a:f:t:j:")) != -1) {
		switch (c) {
			case 'p':
				pluginname=optarg;
				break;
			case 'a':
				algoname=optarg;
				break;
			case 'f':
				patternfile=optarg;
				break;
			case 'j':
				joker=*optarg;
				break;
			case 't':
				type=atoi(optarg);
				break;
			default:
				usage();
		}
	}
	
	/* Checking and parsing arguments */	
	if (!pluginname) usage();
	
	if (patternfile) {
		f=fopen(patternfile,"r");
		if (!f) {
			perror("open(pattern file):");
			usage();
		}
	}
	
	/* Allocating the buffer */	
	buf=malloc(SCANSIZE);
	if (!buf) {
		perror("malloc");
		exit(2);
	}

	/* Loading the plugin */
	printf("Loading plugin '%s'.\n",pluginname);
	plugin=qsearch_load_plugin(pluginname);
	if (!plugin) {
	    fprintf(stderr,"Error: plugin '%s' not found\n", pluginname);
	    exit(3);
	}

	/* Displaying the availables algorithms from the plugin */
	list=qsearch_get_algo_list(plugin);
	if (! *list) {
		fprintf(stderr,"Error plugin '%s' does not provide any algorithms\n",pluginname);
		exit(4);
	}
	printf("Available algorithms in [%s]:\n", pluginname);
	for ( ; *list; list++) {
		printf(" + %s (%s)\n",(*list)->name, (*list)->info);
	}

	/* Getting an algorithm from the plugin */
	if (algoname) {
		algo=qsearch_get_algo(plugin, algoname);
		if (! algo) {
        	        fprintf(stderr,"Error algorithm '%s' not found\n",algoname);
                	exit(4);
		}
	}
	else {
		list=qsearch_get_algo_list(plugin);
		algo=*list; /* Let's take the first one, we know the list is not empty */
	}
	printf("Using algo %s (%s).\n",algo->name,algo->info);

	/* Instanciating the search object */
	options.joker=joker; 
	obj = qsearch_constructor(algo, &options);
	if (! obj) {
	    	fprintf(stderr,"Error building search object with algo '%s'\n",algo->name);
		exit(4);
	}

	/* Adding patterns */
	if (f) {
		char patbuf[1024];
                int ret;
		printf("Adding patterns from file [%s]\n",patternfile);
		while(fgets(patbuf, 1023, f)) {
		        patnb++;
			patbuf[strlen(patbuf)-1]=0;
			if (!strlen(patbuf)) continue;
		        if ((ret=qsearch_add_pattern(obj, patbuf, strlen(patbuf), strdup(patbuf), type))) {
				fprintf(stderr, "Error %i adding pattern [%s]\n",ret,patbuf);
			        exit(5);
			}
                        printf("+%i\t\r",patnb);fflush(stdout);
		}
	}
	else {
                int ret;
		printf("Adding hard-coded pattern:\n");
		patnb=0;
		for (idxp=pat; idxp->ptrn; idxp++) {
		  	if ((ret=qsearch_add_pattern(obj, idxp->ptrn, strlen(idxp->ptrn), idxp->descr, idxp->type))) {
			    	fprintf(stderr, "Error %i adding pattern [%s]. Skipping.\n",ret,idxp->descr);
				continue;
			}
			patnb++;
			printf(" + [%s] type : %i\n",idxp->descr,idxp->type);
		}
	}
		
	/* Compiling the object */
	printf("Compiling the patterns: ");fflush(stdout);
        gettimeofday(&before, NULL);
	if (qsearch_compile(obj)) {
	    	fprintf(stderr,"Error compiling the search object\n");
		exit(6);
	}
        gettimeofday(&after, NULL);
        duration = (double) after.tv_sec-before.tv_sec + (after.tv_usec-before.tv_usec) * 1e-6;
	printf("ok. Compiled %i patterns in %.3fms. (%.1f pat/ms)\n",patnb, 1000*duration, ((double)patnb)/(1000*duration));


	/* Searching for patterns in stdin */
	printf("Searching for patterns in stdin.\n");

	state_in=NULL;
	state_io=NULL;
	ret=0;
	bufofs=0;
	count=0;
	gettimeofday(&before, NULL);

	/* Test weather state_io is supported */
	ret=qsearch_alloc_io_state(obj, &state_io);
	if (ret == -QENOTIMPLEMENTED) { /* not supported */
		printf("State_io not supported, using state_in and state_out\n");
		ret=0;
		while ((ret == 0) && (size=read(0,buf,SCANSIZE))) {			
			ret=qsearch_search(obj, state_in, &state_out, NULL, &cb, &count, buf, size);
			qsearch_free_state(obj, &state_in);
			state_in=state_out;
			bufofs+=size;
			count++;
		}
	}
	else if (ret) {
		fprintf(stderr,"Error %i allocating state_io buffer\n", ret);
		exit(-7);
	}
	else {
		printf("State_io not supported, using state_in and state_out\n");
		while ((ret == 0) && (size=read(0,buf,SCANSIZE))) {			
			ret=qsearch_search(obj, NULL, NULL, state_io, &cb, &count, buf, size);
			bufofs+=size;
			count++;
		}
	}

	gettimeofday(&after, NULL);
	duration = (double) after.tv_sec-before.tv_sec + (after.tv_usec-before.tv_usec) * 1e-6;

	if (ret) {
	    	printf("Search aborted with return value : %i\n", ret);
	}
	printf("Searched %i bytes in %.3fs. (Estimated rate : %.1f Mo/s)\n",bufofs,duration, ((double)bufofs)/(1024*1024*duration));
	
	/* Freeing everything */
	printf("Freeing everything.\n");
	free(buf);
	qsearch_free_state(obj, &state_in);
	qsearch_destructor(obj);
	qsearch_unload_plugin(plugin);
	printf("Done, goodbye.\n");
	return 0;
}


