/*
 * qsearch.c -- qsearch API implementation for the linux kernel.
 *              see http://www.cartel-securite.net/pbiondi/libqsearch.html
 *              for more informations
 *
 * Copyright (C) 2002  Philippe Biondi <biondi@cartel-securite.fr>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/kmod.h>


#define DEBUG
#include <qsearch.h>

/* XXX replace by appropriate code */
#define INCUSECOUNT(plugin)  do {(*((plugin)->add_use_count)(+1));} while(0)
#define DECUSECOUNT(plugin)  do {(*((plugin)->add_use_count)(-1));} while(0)

typedef struct plugin_chain_s plugin_chain_t;

struct plugin_chain_s {
	plugin_chain_t *next;
	qsearch_plugin_t *plugin;
};

plugin_chain_t *chain_root;
static DECLARE_MUTEX(chain_sem);

MODULE_AUTHOR("Philippe Biondi <biondi@cartel_securite.fr>");
MODULE_DESCRIPTION("libqsearch API implementation");
MODULE_LICENSE("GPL");

EXPORT_SYMBOL(qsearch_register);
EXPORT_SYMBOL(qsearch_deregister);

EXPORT_SYMBOL(qsearch_load_plugin);
EXPORT_SYMBOL(qsearch_unload_plugin);
EXPORT_SYMBOL(qsearch_get_algo);
EXPORT_SYMBOL(qsearch_constructor);
EXPORT_SYMBOL(qsearch_add_pattern);
EXPORT_SYMBOL(qsearch_compile);
EXPORT_SYMBOL(qsearch_destructor);
EXPORT_SYMBOL(qsearch_search);
EXPORT_SYMBOL(qsearch_alloc_io_state);
EXPORT_SYMBOL(qsearch_free_state);




/* Get the list of available plugins in QSEARCH_PLUGIN_DIR */
char **qsearch_get_plugin_list(void)
{
        plugin_chain_t *p = chain_root;
	
	DBG("plugin list :\n");
	while (p) {
		DBG(" + [%s]\n", p->plugin->name);
	}
	DBG("end plugin list :\n");
	return NULL;
}
	


/* Free a plugin list allocated by qsearch_get_plugin_list() */
void qsearch_free_plugin_list(char **list);

/* Load a plugin */
qsearch_plugin_t *qsearch_load_plugin(const char *name)
{
       	plugin_chain_t *p;

	DBG("entry\n");

	down(&chain_sem);
       	p = chain_root;
	while (p) {
		if (!strcmp(p->plugin->name, name)) 
			break;
		p = p->next;
	}
	up(&chain_sem);
	if (p) {
		DBG("found plugin [%s] !\n",name);
		(*p->plugin->add_use_count)(+1);
		return  p->plugin;
	}
	else {
		char modname[30];
		DBG("plugin [%s] not loaded, try to load it.\n", name);
		sprintf(modname, "qsearch_%s",name);
		request_module(modname);
		down(&chain_sem);
		p = chain_root;
		while (p) {
			if (!strcmp(p->plugin->name, name)) 
				break;
			p = p->next;
		}
		up(&chain_sem);
		if (p) {
			DBG("found plugin [%s] !\n",name);
			(*p->plugin->add_use_count)(+1);
			return  p->plugin;
		}
		else {
			DBG("plugin [%s] not found\n",name);
			return NULL;
		}
	}
}
		

/* Unload a previously loaded plugin */
int qsearch_unload_plugin(qsearch_plugin_t *plugin)
{
	DBG("begin\n");
	(*plugin->add_use_count)(-1);
	DBG("end\n");
	return 0;
}

/* Get an algo from a plugin */
qsearch_algo_t *qsearch_get_algo(qsearch_plugin_t *plugin, const char *name)
{
	qsearch_algo_t **list;

	DBG("entry\n");
	DBG("looking for algo [%s] in plugin [%s]\n",name, plugin->name);
	list=plugin->algo_list;
	while (*list) {
		if (! strcmp( (*list)->name, name))
			break;
		list++;
	}
	DBG("result: algo at %p\n", *list);
	return *list;
}

/* Get the list of algorithms provided by a plugin */
qsearch_algo_t **qsearch_get_algo_list(qsearch_plugin_t *plugin)
{
	DBG("entry\n");
	return plugin->algo_list;
}

/* Instantiate a search object from a given algorithm */
qsearch_obj_t *qsearch_constructor(qsearch_algo_t *algo, qsearch_options_t *options)
{
	DBG("entry\n");
	return (*algo->constructor)(algo, options);
}

/* Destruct a search object */
int qsearch_destructor(qsearch_obj_t *self)
{
	DBG("entry\n");
	return (*self->statics->destructor)(self);
}

/* Add a search pattern to a search object, and associate pattern data to it */
int qsearch_add_pattern(qsearch_obj_t *self, unsigned char *pattern, int len, void *ptrn_data, int type)
{
	DBG("entry\n");
	if ((1<<type) & self->statics->supported_types)
                return (*self->statics->add_pattern)(self,pattern,len,ptrn_data,type);
        return -QEBADTYPE;
}

/* Once each search pattern has been added to the object, compile it */
int qsearch_compile(qsearch_obj_t *self)
{
	DBG("entry\n");
	return (*self->statics->compile)(self);
}

/* Once the object is compiled, search the buffer, using state_in as start state.
 *  * The very first one state is NULL. Each time a match occurs, the callback function
 *
 */
int qsearch_search(qsearch_obj_t *self,
                   qsearch_state_t *state_in, qsearch_state_t **state_out,
                   qsearch_state_t *state_io,
                   qsearch_callback_t cb, void *cb_data,
                   void *buffer, size_t len)
{
	DBG("entry\n");
	return (*self->statics->search)(self, state_in, state_out, state_io,
                                        cb, cb_data, buffer, len);
}

int qsearch_alloc_io_state(qsearch_obj_t *self, qsearch_state_t **state)
{
	DBG("entry\n");
	return (*self->statics->alloc_io_state)(self, state);
}
int qsearch_free_state(qsearch_obj_t *self, qsearch_state_t **state)
{
	DBG("entry\n");
	return (*self->statics->free_state)(self, state);
}


int qsearch_register(qsearch_plugin_t *plugin)
{
	plugin_chain_t **p, *q; 
	int err = 0;

	DBG("entry\n");

	MOD_INC_USE_COUNT;
	DBG("entry_after\n");
	DBG("plugin=%p\n",plugin);
	DBG("plugin->name=%p\n",plugin->name);

	DBG("want to register plugin [%s]\n", (plugin->name));

	down(&chain_sem);
	p = &chain_root;

	while (*p) {
		if (!strcmp((*p)->plugin->name, plugin->name)) {
			err = -EEXIST;	
			goto error;
		}
		p = &((*p)->next);
	}
	q = kmalloc(sizeof(plugin_chain_t), GFP_KERNEL);
	if (!q) {
		err = -ENOMEM;
		goto error;
	}

	q->next = NULL;
	q->plugin = plugin;
	*p = q;
error:
	up(&chain_sem);
	return err;
}

int qsearch_deregister(qsearch_plugin_t *plugin)
{
	plugin_chain_t **p, *q;
	int err = 0;

        DBG("want to deregister plugin [%s]\n", plugin->name);

	down(&chain_sem);
	p = &chain_root;
	while (*p) {
		if (!strcmp((*p)->plugin->name, plugin->name))
			break;
		p = &((*p)->next);
	}
	if (!*p) {
		err = -ENOENT;
		goto error;
	}
	q = (*p);
	(*p) = (*p)->next;
	kfree(q);
		
error:
	up(&chain_sem);
	if (!err) MOD_DEC_USE_COUNT;
	return err;

}



int __init qsearch_init(void)
{
	chain_root = NULL;
	return 0;
}


#ifdef MODULE
int init_module(void)
{
 	int err;
	printk("Loading qsearch module...\n");
	
	err = qsearch_init();
	if (err)
		goto err_init;

	printk("qsearch module loaded.\n");
	return 0;
err_init:
	printk("Error (%i) loading qsearch module.\n",err);
	return err;
}

void cleanup_module(void)
{
	printk("qsearch module removed\n");
}

#endif

