/*
 * qstest.c -- qsearch test module for kernel API and algos.
 *             see http://www.cartel-securite.net/pbiondi/libqsearch.html
 *             for more informations
 *
 * Copyright (C) 2002  Philippe Biondi <biondi@cartel-securite.fr>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#include <linux/module.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#include <linux/miscdevice.h>
#include <linux/spinlock.h>
#define DEBUG
#include <qsearch.h>

MODULE_AUTHOR("Philippe Biondi <biondi@cartel_securite.fr>");
MODULE_DESCRIPTION("libqsearch API test module");
MODULE_LICENSE("GPL");


#define SHOW_STATE 0
#define LOAD_PLUGIN 1
#define GET_ALGO 2
#define CONSTRUCTOR 3
#define ADD_PATTERN 4
#define COMPILE 5
#define DESTRUCTOR  6
#define UNLOAD_PLUGIN 7


#define QSTEST_MINOR MISC_DYNAMIC_MINOR

#define STATE_UNDEF    0
#define STATE_PLUGIN   1
#define STATE_ALGO     2
#define STATE_CONSTRUCTED 3
#define STATE_COMPILED 4

/*
 *                        UNDEF <---------\
 *                          |             |
 *                          | load_plugin |
 *                          V             | unload_plugin
 *                        PLUGIN --------/|
 *                          |             |
 *                      __  | get_algo    |
 *                     /  V V /-----------/
 *           get_algo |   ALGO <---------\
 *                     \__/ |             |
 *                    __    | constructor |
 *                   /  V   V             |
 *      add_pattern |  CONSTRUCTED ------/| destructor
 *                   \__/   |             |
 *                          | compile     |
 *                          V             |
 *                       COMPILED -------/
 */

static int state;
static int ptrn_nb;
static qsearch_plugin_t *plugin;;
static qsearch_algo_t *algo;
static qsearch_obj_t *obj;
static qsearch_state_t *stin, *stout;
static struct semaphore open_sem;

int cb(void *cb_data, void *ptrn_data, size_t match_offset)
{
	printk(KERN_INFO "XXXX qtest_cb: found pattern %i at %i\n",
	       (int)ptrn_data, match_offset);
	return 0;
}


static ssize_t qstest_write(struct file * file,
                const char * buf, size_t count, loff_t *ppos )
{
	void *kbuf;
	int res;
	DBG("entry\n");

	if (state != STATE_COMPILED)
		return -EBUSY;

	kbuf = kmalloc(count, GFP_KERNEL);
	if (!kbuf)
		return -ENOMEM;
	
	if (copy_from_user(kbuf, buf, count)) {
		kfree(kbuf);
		return -EFAULT;
	}

	DBG("look into [%.10s...]\n",(char *)kbuf);

	res = qsearch_search(obj, stin, &stout, NULL, &cb, NULL, kbuf, count); 
	stin = stout;
	DBG("res = %i\n", res);
	
	kfree(kbuf);
		
	ppos += count;
	return count;
}

static int qstest_open( struct inode *inode, struct file *file )
{
	/* Only one at time */
	if (down_trylock(&open_sem)) {
		return -EAGAIN;
	}
		
	MOD_INC_USE_COUNT;
	return 0;
}

static int qstest_release( struct inode *inode, struct file *file )
{
	up(&open_sem);
	MOD_DEC_USE_COUNT;
	return 0;
}

static int qstest_ioctl(struct inode *inode, struct file *file,
                            unsigned int cmd, unsigned long arg)
{
	char karg[256];
	int err = 0;
	int len;
	DBG("entry, cmd=%i\n", cmd);

	switch (cmd) {
	case SHOW_STATE:
		printk(KERN_DEBUG "qstest: state = %i\n", state);
		printk(KERN_DEBUG "qstest: plugin = %p\n", plugin);
		printk(KERN_DEBUG "qstest: algo = %p\n", algo);
		printk(KERN_DEBUG "qstest: obj = %p\n", obj);
		break;
	case LOAD_PLUGIN:
		if (state != STATE_UNDEF)
			return -EBUSY;
		len = strncpy_from_user((char *)&karg, (char *)arg, 255);
		if (len < 0) return len;
		karg[len]=0;
		DBG("load_plugin: karg=[%s]\n",karg);
		plugin = qsearch_load_plugin(karg);
		if (!plugin)
			return -ENOENT;
		state = STATE_PLUGIN;
		break;
	case GET_ALGO:
		if ((state != STATE_PLUGIN)
                    && (state != STATE_ALGO))
			return -EBUSY;
		len = strncpy_from_user((char *)&karg, (char *)arg, 255);	
		if (len < 0) return len;
		karg[len]=0;
		DBG("get_algo: karg=[%.256s]\n",karg);
		algo = qsearch_get_algo(plugin, karg);
		if (!algo)
			return -ENOENT;
		state = STATE_ALGO;
		break;
	case CONSTRUCTOR:
		if (state != STATE_ALGO)
			return -EBUSY;
		obj = qsearch_constructor(algo, NULL);
		if (!obj)
			return -EIO;
		ptrn_nb = 0;
		state = STATE_CONSTRUCTED;
		break;
	case ADD_PATTERN: /* XXX need better support with a struct as arg */
		if (state != STATE_CONSTRUCTED)
			return -EBUSY;
		len = strncpy_from_user((char *)karg, (char *)arg, 255);	
		if (len < 0) return len;
		karg[len]=0;
		err = qsearch_add_pattern(obj, karg, strlen(karg), (void *)ptrn_nb, 0);
		if (err)
			return -EIO;
		printk(KERN_INFO "add_pattern: pattern %i = [%s]\n", ptrn_nb, karg);
		ptrn_nb++;
		break;
	case COMPILE:
		if (state != STATE_CONSTRUCTED)
			return -EBUSY;
		err = qsearch_compile(obj);
		if (err) return -EIO;
		if (stin)
			qsearch_free_state(obj, &stin);
		stin = NULL;
		state = STATE_COMPILED;
		break;
	case DESTRUCTOR:
		if ((state != STATE_CONSTRUCTED) 
		    && (state != STATE_COMPILED))
			return -EBUSY;
		err = qsearch_destructor(obj);
		if (err)
			return -EIO;
		obj = NULL;
		state = STATE_ALGO;
		break;
	case UNLOAD_PLUGIN:
		if ((state != STATE_ALGO)
		    && (state != STATE_PLUGIN))
			return -EBUSY;
		err = qsearch_unload_plugin(plugin);
		if (err) return -EIO;
		plugin = NULL;
		state = STATE_UNDEF;
		break;
	default:
		return -EINVAL;
	}	
	return 0;
}

static struct file_operations qstest_fops = {
	owner: 		THIS_MODULE,
        write: 		qstest_write,
        ioctl: 		qstest_ioctl,
        open: 		qstest_open,
        release: 	qstest_release,
};


static struct miscdevice qstest_dev = {
	QSTEST_MINOR,
	"qstest",
	&qstest_fops
};

static int __init qstest_init(void)
{
	int ret = 0;

	state = STATE_UNDEF;
	init_MUTEX(&open_sem);

	ret = misc_register( &qstest_dev );
	if (ret) {
		printk(KERN_ERR "qstest: can't misc_register on minor=%d\n", 
		       QSTEST_MINOR);
		goto err_misc_reg;
	}
	printk(KERN_INFO "qstest loaded with minor=%d\n",qstest_dev.minor);
	goto no_err;

/*err_other:*/
	misc_deregister( &qstest_dev );
err_misc_reg:
no_err:
	return ret;
}

static void __exit qstest_cleanup_module (void)
{
	misc_deregister( &qstest_dev );
	if (obj)
		qsearch_destructor(obj);
	if (plugin)
		qsearch_unload_plugin(plugin);
	printk(KERN_INFO "qstest module removed\n");
}

module_init(qstest_init);
module_exit(qstest_cleanup_module);


