/*
 * User-space program to setup the qstest module
 *
 * Copyright (C) 2002  Arkoon Network Security
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 *
 * How to use this program:
 * 1. Compile and load the qsearch* and qstest modules
 * 2. Create the qstest device (mknod /dev/qstest c 10 63)
 * 3. Load a plugin, an algo, construct and add patterns,
 *    using commands 1, 2, 3, 4 and 5 of this program
 * 4. Send your data to /dev/qstest; if anything matches,
 *    a message will appear in the kernel logs.
 */

#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h> /* getopt */
#include <stdlib.h> /* atoi */

#define QSTEST_DEV    "/dev/qstest"

/* From qstest.c
 *
 *
 *                        UNDEF <---------\
 *                          |             |
 *                          | load_plugin |
 *                          V             | unload_plugin
 *                        PLUGIN --------/|
 *                          |             |
 *                      __  | get_algo    |
 *                     /  V V /-----------/
 *           get_algo |   ALGO <---------\
 *                     \__/ |             |
 *                    __    | constructor |
 *                   /  V   V             |
 *      add_pattern |  CONSTRUCTED ------/| destructor
 *                   \__/   |             |
 *                          | compile     |
 *                          V             |
 *                       COMPILED -------/
 *
 */

#define SHOW_STATE    0
#define LOAD_PLUGIN   1
#define GET_ALGO      2
#define CONSTRUCTOR   3
#define ADD_PATTERN   4
#define COMPILE       5
#define DESTRUCTOR    6
#define UNLOAD_PLUGIN 7

#define IOCTL(cmd, data) qstest_ioctl(cmd, data) >= 0           \
                        ? fprintf(stdout, " ok.\n")             \
                        : fprintf(stdout, " failed.\n")

static void usage() {
        fprintf(stderr, "Usage: qstest -c <command> [ -d <data> ]\n"
                "  Available commands:\n"
                "   0 (SHOW_STATE)\n"
                "   1 (LOAD_PLUGIN), data: plugin name\n"
                "   2 (GET_ALGO), data: algo name\n"
                "   3 (CONSTRUCTOR)\n"
                "   4 (ADD_PATTERN), data: pattern\n"
                "   5 (COMPILE)\n"
                "   6 (DESTRUCTOR)\n"
                "   7 (UNLOAD_PLUGIN)\n");
        exit(1);
}

int qstest_ioctl(unsigned int cmd,
                 char *data)
{
        int rc, fd;
     
        fd = open(QSTEST_DEV, O_RDWR);
     
        if(fd < 0)
        {
                fprintf(stderr, "qstest_ioctl: Couldn't open qstest device.\n");
                return -1;
        }
     
        rc = ioctl(fd, cmd, data);
     
        if(rc < 0)
        {
                fprintf(stderr, "qstest_ioctl: ioctl failed.\n");
        }
        
        close(fd);
     
        return rc;
}

int main(int argc, char **argv)
{
        int opt, cmd = -1;
        char *data = NULL;

        /* Parse the command line */
        while((opt = getopt(argc, argv, "c:d:")) != -1) {
                switch(opt) {
                case 'c':
                        cmd = atoi(optarg);
                        break;
                case 'd':
                        data = optarg;
                        break;
                default:
                        usage();
                }
        }

        /* No args? */
        if(cmd == -1) usage();
     
        /* let's go... */
        if(!data) { /* just a command */
                switch(cmd) {
                case SHOW_STATE:
                        fprintf(stdout, "Showing the state of the module...");
                        IOCTL(cmd, data);
                        break;
                case CONSTRUCTOR:
                        fprintf(stdout, "Calling the constructor...");
                        IOCTL(cmd, data);
                        break;
                case DESTRUCTOR:
                        fprintf(stdout, "Calling the destructor...");
                        IOCTL(cmd, data);
                        break;
                case COMPILE:
                        fprintf(stdout, "Compiling the patterns...");
                        IOCTL(cmd, data);
                        break;
                default:
                        fprintf(stderr,
                                "This command (%d) isn't valid here, does it require data?\n", cmd);
                        exit(1);
                }
        }
        else { /* command and data */
                if(strlen(data) > 255) usage();
                switch(cmd) {
                case LOAD_PLUGIN:
                        fprintf(stdout, "Loading plugin: \"%s\"...", data);
                        IOCTL(cmd, data);
                        break;
                case GET_ALGO:
                        fprintf(stdout, "Loading algo: \"%s\"...", data);
                        IOCTL(cmd, data);
                        break;
                case ADD_PATTERN:
                        fprintf(stdout, "Adding pattern: \"%s\"...", data);
                        IOCTL(cmd, data);
                        break;
                case UNLOAD_PLUGIN:
                        fprintf(stdout, "Unloading plugin: \"%s\"...", data);
                        IOCTL(cmd, data);
                        break;
                default:
                        fprintf(stderr, "This command (%d) isn't valid here (not a data command?)...\n", cmd);
                        exit(1);
                }
        }
        
        return 0;     
}
