/*
 * bm.c -- A kind of Boyer-Moore algoithm for the qsearch API.
 *         see http://www.cartel-securite.net/pbiondi/libqsearch.html
 *         for more informations
 *
 * Copyright (C) 2002  Philippe Biondi <biondi@cartel-securite.fr>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */


/* qsearch.h is the only authorized (and mandatory) 
 * header file. It will include eveything we need
 * wether we are in user space or not.
 * In user space, we need:
 * #include <errno.h>
 * #include <stdlib.h>
 * #include <ctype.h>
 * #include <string.h>
 */

#include "qsearch.h"


typedef unsigned char uchar;

typedef struct pattern_tuple {
    	size_t len;
	int type;
    	void *ptrn_data;
    	uchar pattern[0];
} tuple_t;
    
typedef struct pattern_list {
	struct pattern_list *next;
	struct pattern_tuple *tuple_t;
} pelt_t;
      

typedef struct test_and_skip {
	int shift;
	int testnb;
	tuple_t *test[1];
} tns_t;

typedef struct bm_algo_s bm_algo_t;
typedef struct bm_obj_s  bm_obj_t;
typedef struct bm_options_s bm_options_t;
typedef struct bm_state_s bm_state_t;

struct bm_state_s {
        size_t done;
	int skip;
        int len;
        uchar remainder[0];
};

struct bm_algo_s {
        QSEARCH_ALGO(bm)
};

struct bm_options_s {
        QSEARCH_OPTIONS(bm)
};

struct bm_obj_s {
	QSEARCH_OBJ(bm);
	tns_t *skip[256];
	unsigned int min,max,nb;
	pelt_t *list_head;	
};

typedef int (*memcmp_t)(const void *s1, const void *s2, size_t n);

#define MIN(x,y) ((x)<(y)) ? (x) : (y)

/*
 * Here are specialized memcmp() 
 */

static int memcmp_cs(const void *s1, const void *s2, size_t n) 
{
        while (n--) {
                if ((*(const uchar *)(s1++))-(*(const uchar *)(s2++)))
			return 1;
        }
	return 0;
}

static int memcmp_ci(const void *s1, const void *s2, size_t n) 
{
        /* we assume that s2 is already lowercase */
        while (n--) {
                if (tolower(*(const uchar *)(s1++))-(*(const uchar *)(s2++))) 
			return 1;
        }
	return 0;
}

static int memcmp_joker(const void *s1, const void *s2, size_t n)
{
        while (n--) {
                if ( *(const uchar *)(s2) == '*' ) {
                        s1++; s2++;
                        continue;
                }
                if ((*(const uchar *)(s1++))-(*(const uchar *)(s2++))) {
			return 1;
                }
        }
        return 0;
}

static int memcmp_joker_ci(const void *s1, const void *s2, size_t n)
{
        /* we assume that s2 is already lowercase */
        while (n--) {
                if ( *(const uchar *)(s2) == '*' ) {
                        s1++; s2++;
                        continue;
                }
                if (tolower(*(const uchar *)(s1++))-(*(const uchar *)(s2++))) {
			return 1;
                }
        }
        return 0;
}



static memcmp_t memcmps[] = {
		&memcmp_cs,
		&memcmp_ci,
		&memcmp_joker,
		&memcmp_joker_ci,
};

/*
 * Constructor
 */

static bm_obj_t *bm_constructor(bm_algo_t *algo, bm_options_t *options)
{
	bm_obj_t *self;

	self = malloc(sizeof(bm_obj_t));
	if (! self) 
		return NULL;
	self->statics=algo;

	memset(self->skip, 0, sizeof(self->skip));
	self->min = ~0;
	self->max = 0;

	return self;
}

/*
 * Destructor
 */

static int bm_destructor(bm_obj_t *self)
{	
	tns_t **p;
	tuple_t **q;
	int i,j;

	for (i=0, p=self->skip; i < 256; i++, p++) {
		if (*p) {
			for (j=0, q=(*p)->test; j < (*p)->testnb ; j++, q++) {
				if ((*q)->type >= 0)
					(*q)->type = -1;
				else
					*q = NULL;
			}
		}
	}

					
	
	for (i=0, p=self->skip; i < 256; i++, p++) {
		if (*p) {
			q = (*p)->test;
			for (j=0, q=(*p)->test; j < (*p)->testnb ; j++, q++) {
				if (*q) free(*q);
			}
			free(*p);
		}
	}
	
	free(self);
	return 0;
}

/*
 * Add pattern
 */

static int add_shift(bm_obj_t *obj, unsigned char letter, int pos)
{
	tns_t *s;

	if ((s = obj->skip[letter])) {
		if (s->shift > pos) 
			s->shift = pos;
	}
	else {
		s = malloc(sizeof(tns_t));
		if (!s)
			return -2;
		obj->skip[letter] = s;
		s->shift = pos;
		s->testnb = 0;
		s->test[0] = NULL;
	}
	return 0;
}

static int add_last_shift(bm_obj_t *obj, unsigned char letter, tuple_t *tuple)
{
	tns_t *s;
	if ((s=obj->skip[letter])) {
		s=realloc(s,sizeof(tns_t)+(1+s->testnb)*sizeof(tuple_t *));
		if (!s)
			return -1;
		if (s->shift > tuple->len) 
			s->shift = tuple->len;
		s->test[s->testnb] = tuple;
		s->testnb++;
		s->test[s->testnb] = NULL;
	}
	else {
		s=malloc(sizeof(tns_t)+sizeof(tuple_t *));
		if (!s)
			return -2;
		s->shift = tuple->len;
		s->testnb = 1;
		s->test[0] = tuple;
		s->test[1] = NULL;
	}
	obj->skip[letter] = s;
	return 0;
}


static int bm_add_pattern(bm_obj_t *self, unsigned char *pattern, int len, void *ptrn_data, int type)
{	
	unsigned char *p, c;
	int i, ret;
	tuple_t *tuple;

	if (len <= 0) return 0;

	self->nb++;
	if (self->min > len) 
		self->min = len;
	if (self->max < len)
		self->max = len;

	tuple=malloc(sizeof(tuple_t)+len);
	if (!tuple)
		goto err_malloc1;	
	tuple->len = len;
	tuple->type = type;
	tuple->ptrn_data = ptrn_data;
	switch (type) {
		case QSEARCH_CS:
		case QSEARCH_JOKER_CS:
			memcpy(tuple->pattern,pattern,len);
			break;
		case QSEARCH_CI:
        	case QSEARCH_JOKER_CI:
			for (i=0; i<len; i++) 
				tuple->pattern[i] = tolower(pattern[i]);
			break;
		default:
			return -QEINTERNAL;
	}
	
	for (i=len-1, p=tuple->pattern; i > 0; i--, p++) {
		switch(type) {
		case QSEARCH_CS:
			if (add_shift(self, *p, i)) 
				goto err_malloc2;
			break;
		case QSEARCH_CI:
			if (add_shift(self, *p, i)) goto err_malloc2;
			c = toupper(*p);
			if (c == *p) break;
			if (add_shift(self, c, i)) goto err_malloc2;
			break;
		case QSEARCH_JOKER_CS:
			if (*p == '*') {
				c=0;
				do {
					if (add_shift(self, c, i)) goto err_malloc2;
					c++;
				} while (c);
			} 
			else {
				if (add_shift(self, *p, i)) goto err_malloc2;
			}
			break;
		case QSEARCH_JOKER_CI:
			if (*p == '*') {
				c=0;
				do {
					if (add_shift(self, c, i)) goto err_malloc2;
					c++;
				} while (c);
			} 
			else {
				if (add_shift(self, *p, i)) goto err_malloc2;
				c = toupper(*p);
				if (c == *p) break;
				if (add_shift(self, c, i)) goto err_malloc2;
			}
			break;
		default:
			return -QEINTERNAL;
		}

	}


	switch(type) {
	case QSEARCH_CS:
		ret=add_last_shift(self, *p, tuple);
		break;
	case QSEARCH_CI:
		if ((ret=add_last_shift(self, *p, tuple))) break;
		c = toupper(*p);
		if (c == *p) break;
		ret=add_last_shift(self, c, tuple);
		break;
	case QSEARCH_JOKER_CS:
		if (*p == '*') {
			c=0;
			do {
				if ((ret=add_last_shift(self, c, tuple))) break;
				c++;
			} while (c);
		} 
		else {
			ret=add_last_shift(self, *p, tuple);
		}
		break;
	case QSEARCH_JOKER_CI:
		if (*p == '*') { /* I hope nobody did that ! */
			c=0;
			do {
				if ((ret=add_last_shift(self, c, tuple))) break;
				c++;
			} while (c);
		} 
		else {
			if ((ret=add_last_shift(self, *p, tuple))) break;
			c = toupper(*p);
			if (c == *p) break;
			ret=add_last_shift(self, c, tuple);
		}
		break;
	default:
		return -QEINTERNAL;
	}


	if (ret == -1) goto err_malloc1;
	if (ret) goto err_malloc2;

	return 0;

err_malloc2:
	free(tuple);
err_malloc1:
	return -QENOMEM;
}

/*
 * Compile
 */
static int bm_compile(bm_obj_t *self)
{
	tns_t **p;
	int i;

	for (i=0, p=self->skip; i < 256; i++, p++) {
		if (*p) {
			if ((*p)->shift > self->min)
				(*p)->shift = self->min;
		}
		else {
			(*p) = malloc(sizeof(tns_t));
			if (!*p) 
				return -QENOMEM;
			(*p)->shift = self->min;
			(*p)->testnb = 0;
			(*p)->test[0] = NULL;
		}
	}	
	
	return 0;
}

static int bm_search(bm_obj_t *self, 
		     bm_state_t *state_in, bm_state_t **state_out,
		     bm_state_t *state_io,
		     qsearch_callback_t cb, void *cb_data,
		     void *buffer, size_t len)
{
	size_t pos,begin=0;
	tns_t *s;
	tuple_t **q;
	int ret;
	memcmp_t mcmp;

	pos = self->min;

	/*
	 * Searching the input state (if present)
	 */
	if (state_in || (state_in=state_io)) {
		begin = state_in->done;
		pos = state_in->skip;
		while ((pos < self->max) && (pos <= len)) {
			s = self->skip[*((uchar *)buffer+pos-1)];
			if (!s)
				return -QEINTERNAL;
			for(q=s->test; (*q); q++) {
				mcmp = memcmps[(*q)->type];
				if (pos < (*q)->len) {
					if (state_in->len+pos < (*q)->len)
						continue;
					if ((*mcmp)(buffer, (*q)->pattern+(*q)->len-pos, pos))
						continue;
					if ((*mcmp)(state_in->remainder+state_in->len+pos-(*q)->len, (*q)->pattern, (*q)->len-pos))
						continue;
				}
				else {
					if ((*mcmp)(buffer+pos-(*q)->len, (*q)->pattern, (*q)->len))
						continue;
				}
				ret = (*cb)(cb_data, (*q)->ptrn_data, begin+pos);
				if (ret)
					return ret;
			}
			pos += s->shift;
		}
	}
	    
	/*
	 * Search in the buffer
	 */	
        while (pos <= len) {
		s = self->skip[*((uchar *)buffer+pos-1)];
		if (!s) 
			return -QEINTERNAL;
		for(q=s->test; (*q); q++) {
			if (pos < (*q)->len)
				continue;
			mcmp = memcmps[(*q)->type];
			if ((*mcmp)(buffer+pos-(*q)->len, (*q)->pattern, (*q)->len))
				continue;
			ret = (*cb)(cb_data, (*q)->ptrn_data, begin+pos);
			if (ret)
				return ret;
		}
		pos += s->shift;
	}

	/*
	 * Creating the output state
	 */

	if (state_out || (state_io && (state_out=&state_io))) {
		/*       in remainder  buffer
		 *      <-----a-----> <---b--->
	 	 *   <-----------c------------>
		 *           out remainder
		 */
		size_t C, c = self->max-1;
		size_t B, b = len;
		size_t A, a = 0;
		if (state_in)
			a = state_in->len;

		C = MIN(c, b+a);
		B = MIN(c, b);
		A = C-B;
		

		if (state_out != &state_io) {
			(*state_out) = malloc(sizeof(struct bm_state_s)+C);
			if (!(*state_out)) return -QENOMEM;
		}
		if (A) {
			if (*state_out == state_in) {
				memmove((*state_out)->remainder, state_in->remainder+a-A, A);
			}
			else {
				memcpy((*state_out)->remainder, state_in->remainder+a-A, A);
			}
		}


		if (B)
			memcpy((*state_out)->remainder+A, buffer+b-B, B);
		(*state_out)->len = C;
		(*state_out)->done = begin+len;
		(*state_out)->skip = pos-len;
	}
				
	return 0;
}

/*
 * Alloc state IO
 */
static int bm_alloc_io_state(bm_obj_t *self, bm_state_t **state)
{
	*state = malloc(sizeof(bm_state_t)+self->max);
	if (!*state) return -QENOMEM;
	(*state)->len = 0;
	(*state)->done = 0;
	(*state)->skip = 0;
	return 0;
}

/*
 * Free states
 */
static int bm_free_state(bm_obj_t *self, bm_state_t **state)
{
	if (*state) {
		free(*state);
		*state = NULL;
	}	
  	return 0;
}

/*
 * Structures
 */
static bm_algo_t algo_bm = {
	"bm",
	"Boyer-Moore, extended to pattern sets",
	1<<QSEARCH_CS | 1<<QSEARCH_CI | 1<<QSEARCH_JOKER_CS | 1<<QSEARCH_JOKER_CI,
	&bm_constructor,
	&bm_destructor,
	&bm_add_pattern,
	&bm_compile,
	&bm_search,
	&bm_alloc_io_state,
	&bm_free_state,
};

bm_algo_t *QSEARCH_ALGO_LIST_SYM[] = {
	&algo_bm,
	NULL
};


int QSEARCH_API_VERSION_SYM = QSEARCH_API_VERSION;
