/*
 * simple.c -- a simple search algorithm for the qsearch API
 *             see http://www.cartel-securite.net/pbiondi/libqsearch.html
 *             for more informations
 *
 * Copyright (C) 2002  Philippe Biondi <biondi@cartel-securite.fr>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

/* qsearch.h is the only authorized (and mandatory) 
 * header file. It will include eveything we need
 *  wether we are in user space or not.
 */

#include "qsearch.h"

typedef unsigned char uchar;

struct pattern_tuple {
    	size_t len;
	int type;
    	void *ptrn_data;
    	uchar pattern[0];
};
    
struct pattern_list {
	struct pattern_list *next;
	struct pattern_tuple *tuple;
};
      

typedef struct simple_state_s simple_state_t;
typedef struct simple_algo_s simple_algo_t;
typedef struct simple_obj_s simple_obj_t;
typedef struct simple_options_s simple_options_t;

struct simple_options_s {
	QSEARCH_OPTIONS(simple);
};

struct simple_algo_s {
        QSEARCH_ALGO(simple);
};


/* Case sensitive/insensitive or joker optimizations */

#define TYPE_CS 0
#define TYPE_CI 1
#define TYPE_JOKER 2
#define TYPE_JOKER_CI 3

#define NB_OF_TYPES 4


struct simple_obj_s {
	QSEARCH_OBJ(simple);
	struct pattern_list *list_head;	
	int pattern_nb;
	char joker;
	unsigned int min,max;
	int size[NB_OF_TYPES*256];
	struct pattern_tuple ***patterns;
};

struct simple_state_s {
	size_t pos;
    	size_t len;
	uchar remainder[1];
};


static int memcmp_cs(const void *s1, const void *s2, size_t n, simple_obj_t *obj) 
{
        while (n--) {
                if ((*(const uchar *)(s1++))-(*(const uchar *)(s2++)))
			return 1;
        }
	return 0;
}


static int memcmp_ci(const void *s1, const void *s2, size_t n, simple_obj_t *obj) 
{
	/* we assume that s2 is already lowercase */
	while (n--) {
		if (tolower(*(const uchar *)(s1++))-(*(const uchar *)(s2++))) {
				return 1;
		}
	}
	return 0;
}

static int memcmp_joker(const void *s1, const void *s2, size_t n, simple_obj_t *obj) 
{
	while (n--) {
	    	if ( *(const uchar *)(s2) == obj->joker ) {
		    	s1++; s2++;
			continue;
		}
		if ((*(const uchar *)(s1++))-(*(const uchar *)(s2++))) {
				return 1;
		}
	}
	return 0;
}

static int memcmp_joker_ci(const void *s1, const void *s2, size_t n, simple_obj_t *obj) 
{
	/* we assume that s2 is already lowercase */
	while (n--) {
	    	if ( *(const uchar *)(s2) == obj->joker ) {
		    	s1++; s2++;
			continue;
		}
		if (tolower(*(const uchar *)(s1++))-(*(const uchar *)(s2++))) {
				return 1;
		}
	}
	return 0;
}


static int (*memcmp_array[])(const void *s1, const void *s2, size_t n, simple_obj_t *obj) = {
	&memcmp_cs,
	&memcmp_ci,
	&memcmp_joker,
	&memcmp_joker_ci
};


static simple_obj_t *simple_constructor(simple_algo_t *algo, simple_options_t *options)
{
	simple_obj_t *self;

	self=malloc(sizeof(simple_obj_t));
	if (! self) goto cerr_malloc1;
	self->statics=algo;

	if (options) self->joker = options->joker;
	else self->joker = '*';
	self->pattern_nb=0;
	self->max=0;
	self->min=-1;
	self->list_head=NULL;
	self->patterns=malloc(NB_OF_TYPES*256*sizeof(struct pattern_tuple **));
	if (!self->patterns) goto cerr_malloc2;
	memset(self->patterns,0,NB_OF_TYPES*256*sizeof(struct pattern_tuple **));
	memset(self->size,0,NB_OF_TYPES*256*sizeof(int));

	return self;
cerr_malloc2:
	free(self);
cerr_malloc1:
	return NULL;
}

static int simple_destructor(simple_obj_t *self)
{
  	struct pattern_list *p,*q;
	struct pattern_tuple ***r,**s;
	int i,j;
	

	/* First, mark patterns to avoid multiple free() */
	r=self->patterns;
	for (i=0 ; i < NB_OF_TYPES*256 ; i++, r++) {
		s=*r;
		for (j=0 ; j < self->size[i] ; j++ , s++) {
			if ( (*s)->type != -1) {
				(*s)->type=-1;
			}
			else {
				(*s)=NULL;
			}
		}
	}

	/* Free the list */
	p=self->list_head;
	while (p) {
	    	q=p;
		p=p->next;
		if (q->tuple->type != -1) free(q->tuple);
		free(q);
	}


	/* Once we have one and only one pointer on each pattern, we can free them */
	r=self->patterns;
	for (i=0 ; i < NB_OF_TYPES*256 ; i++, r++) {
		if ((s=*r)) {
			for (j=0 ; j < self->size[i] ; j++ , s++) {
				if (*s) free(*s);
			}
			free(*r);
		}
	}
	free(self->patterns);
  		
	free(self);
	return 0;
}

static int simple_add_pattern(simple_obj_t *self, uchar *pattern, int len, void *ptrn_data, int type)
{
  	struct pattern_tuple *p;
	struct pattern_list *r;
	int i;

	if (len <= 0) return 0;

	if (len > self->max) self->max=len;
	if ((unsigned int)len < self->min) self->min=len;

	/* Create a new tuple */
	p=malloc(sizeof(struct pattern_tuple)+len);
	if (!p) goto aperr_malloc1;
	p->len=len;
	p->ptrn_data=ptrn_data;
	switch (type) {
	case QSEARCH_CS:
		p->type=TYPE_CS;
		self->size[TYPE_CS*256+pattern[0]]++;
		memcpy(p->pattern,pattern,len);
		break;
	case QSEARCH_JOKER_CS:
		p->type=TYPE_JOKER;
		memcpy(p->pattern,pattern,len);
		if (pattern[0] == self->joker)
			for (i=0; i < 256; i++) {
				self->size[TYPE_JOKER*256+i]++;
			}
		else
			self->size[TYPE_JOKER*256+pattern[0]]++;
		break;
	case QSEARCH_CI:
		p->type=TYPE_CI;
		self->size[TYPE_CI*256+toupper(pattern[0])]++;
		if (toupper(pattern[0]) != tolower(pattern[0]))
			self->size[TYPE_CI*256+tolower(pattern[0])]++;
		for (i=0; i<len; i++) {
			p->pattern[i]=tolower(pattern[i]);
		}
		break;
	case QSEARCH_JOKER_CI:
		p->type=TYPE_JOKER_CI;
		if (pattern[0] == self->joker)
			for (i=0; i < 256; i++) {
				self->size[TYPE_JOKER_CI*256+i]++;
			}
		else {			
			self->size[TYPE_JOKER_CI*256+toupper(pattern[0])]++;
			if (toupper(pattern[0]) != tolower(pattern[0]))
				self->size[TYPE_JOKER_CI*256+tolower(pattern[0])]++;
		}
		for (i=0; i<len; i++) {
			p->pattern[i]=tolower(pattern[i]);
		}
		break;
	default:
		free(p);
		return -QEINTERNAL;
	}
		
	/* Create a new list leaf and insert it */
	r=malloc(sizeof(struct pattern_list));
	if (!r) goto aperr_malloc2;
	r->tuple=p;
	r->next=self->list_head;
	self->list_head=r;

	self->pattern_nb++;

	return 0;

aperr_malloc2:
	free(p);
aperr_malloc1:
	return -QENOMEM;
}

static void insert_tuple(simple_obj_t *self, struct pattern_tuple *tuple, uchar place)
{
	struct pattern_tuple **s;
	s=self->patterns[tuple->type*256+place];
	while (*s) s++;
	*s=tuple;
}

	

/* Change the linked list into an array */
static int simple_compile(simple_obj_t *self)
{
	struct pattern_tuple ***p;
	struct pattern_list *q,*r;
	int i;
	int err=0;

	p=self->patterns;
	for (i=0; i < 256*NB_OF_TYPES; i++, p++) {
		if (self->size[i]) {

			*p=malloc((self->size[i]+1)*sizeof(struct pattern_tuple *));
			if (!*p) {
				err=-ENOMEM;
				break;
			}
			memset(*p,0,(self->size[i]+1)*sizeof(struct pattern_tuple *));
		}
	}
	p=self->patterns;
	if (err) {
		for (i=0; i < 256*NB_OF_TYPES; i++, p++) {
			if (*p) free(p);
		}
		return err;
	}

	      
	q=self->list_head;
	self->list_head=NULL;
	while (q) {
	    r=q;
	    q=q->next;

	    switch (r->tuple->type) {
	    case TYPE_CS:
		    insert_tuple(self, r->tuple, r->tuple->pattern[0]);
		    break;
	    case TYPE_JOKER:
		    if (r->tuple->pattern[0] == self->joker) {
			    for (i=0; i < 256 ; i++) 
				    insert_tuple(self, r->tuple, i);
		    }
		    else 
			    insert_tuple(self, r->tuple, r->tuple->pattern[0]);
		    break;
	    case TYPE_CI:
		    insert_tuple(self, r->tuple, r->tuple->pattern[0]);
		    if (toupper(r->tuple->pattern[0]) != r->tuple->pattern[0]) 
			    insert_tuple(self, r->tuple, toupper(r->tuple->pattern[0]));
		    break;
	    case TYPE_JOKER_CI:
		    if (r->tuple->pattern[0] == self->joker) {
			    for (i=0; i < 256 ; i++) 
				    insert_tuple(self, r->tuple, i);
		    }
		    else {
			    insert_tuple(self, r->tuple, r->tuple->pattern[0]);
			    if (toupper(r->tuple->pattern[0]) != r->tuple->pattern[0]) 
				    insert_tuple(self, r->tuple, toupper(r->tuple->pattern[0]));
		    }
		    break;
	    default:
		    p=self->patterns;
		    if (err) {
			    for (i=0; i < 256*NB_OF_TYPES; i++, p++) {
				    if (*p) free(p);
			    }
			    return -QEINTERNAL;
		    }
	    }
	    p++;
	    free(r);
	}
	*p=NULL;
  
	return 0;
}

#ifdef DEBUG
static uchar pattern[100];

static uchar *ptrn(uchar *buf, int len)
{
  	memcpy(pattern,buf,len);
	pattern[len]=0;
	return pattern;
}

static void dump_state(simple_state_t state)
{
  	int i;
	printf("## %i |",state->len);	
  	for (i=0; i < state->len; i++) {
	    	printf("%02x ",*(&state->remainder+i));
	}
	printf("\b| ##\n");
}
#endif 

static int simple_search(simple_obj_t *self, 
			   simple_state_t *state_in, simple_state_t **state_out,
			   simple_state_t *state_io, 
			   qsearch_callback_t cb, void *cb_data,
			   void *buffer, size_t len)
{	
  	size_t ofs, begin=0;
	struct pattern_tuple **p;
	int ret;
	int i;

	/* search in the remainder of the latest search we got from state_in */
  	if (state_in) {
		begin=state_in->pos;
		for (ofs=0; ofs < state_in->len; ofs++) {
			for (i=0; i < NB_OF_TYPES; i++) {
				p=self->patterns[256*i+state_in->remainder[ofs]];
				if (!p) continue;
				for (; *p; p++) {
					if ((state_in->len >= (*p)->len+ofs) || ((*p)->len+ofs-state_in->len > len)) 
						continue;	
					if ((*memcmp_array[(*p)->type])(state_in->remainder+ofs, (*p)->pattern, state_in->len-ofs, self)) 
						continue;
					if ((*memcmp_array[(*p)->type])(buffer, (*p)->pattern+state_in->len-ofs,
											 (*p)->len+ofs-state_in->len, self))
						continue;
					ret=(*cb)(cb_data, (*p)->ptrn_data, begin+ofs+(*p)->len-state_in->len);
					if (ret)
					       	return ret;
				}
			}
		}
	}
	
	/* search in the part where patterns fit in size (FIXME: could be optimized by sorting patterns) */
	for (ofs=0; ofs+self->max < len; ofs++) {
		for (i=0; i < NB_OF_TYPES; i++) {
			p=self->patterns[256*i+*((uchar *)buffer+ofs)];
		       	if (!p) continue;
			for (; *p; p++) {
				if ((*memcmp_array[(*p)->type])(buffer+ofs, (*p)->pattern, (*p)->len, self))
				       continue;	
				ret=(*cb)(cb_data, (*p)->ptrn_data, begin+ofs+(*p)->len);
				if (ret)
					return ret;
			}
		}
	}

	/* search in the end, only for patterns that fit (FIXME: same as above) */
	for (; ofs < len; ofs++) {
		for (i=0; i < NB_OF_TYPES; i++) {
			p=self->patterns[256*i+*((uchar *)buffer+ofs)];
			if (!p) continue;
		       	for (; *p; p++) {
				if (len < (*p)->len+ofs) 
					continue;
				if ((*memcmp_array[(*p)->type])(buffer+ofs, (*p)->pattern, (*p)->len, self))
					continue;
				ret=(*cb)(cb_data, (*p)->ptrn_data, begin+ofs+(*p)->len);
				if (ret)
					return ret;
			}
		}
	}

	/* Put the remainder in state_out */

	if (state_out) {
		(*state_out)=malloc(sizeof(struct simple_state_s)+self->max);
		if (!(*state_out)) return -QENOMEM;
		(*state_out)->pos=begin+len;
		if (len > self->max-1) {
			(*state_out)->len=self->max-1;
			memcpy((*state_out)->remainder, buffer+len-self->max+1, self->max-1);
		}
		else {
			(*state_out)->len=0;
		    	if (state_in) {
			    	if (state_in->len > self->max-1-len) {
				    	memcpy((*state_out)->remainder, state_in->remainder+state_in->len+1+len-self->max, self->max-1-len);
					(*state_out)->len=self->max-1-len;
				}
				else {
	                                 memcpy((*state_out)->remainder, state_in->remainder, state_in->len);
					 (*state_out)->len=state_in->len;
				}
			}
			memcpy((*state_out)->remainder+(*state_out)->len,buffer,len);
			(*state_out)->len += len;
		}
#ifdef DEBUG
	/*	dump_state(*state_out);*/
#endif
	}
	return 0;
}

static int simple_alloc_io_state(simple_obj_t *self, simple_state_t **state)
{
	return -QENOTIMPLEMENTED;
}

static int simple_free_state(simple_obj_t *self, simple_state_t **state)
{
  	if (*state) {
	  	free(*state);
		*state=NULL;
	}
  	return 0;
}



static simple_algo_t algo_simple = {
	"simple",
	"This is a simple search algo",
	1<<QSEARCH_CS | 1<<QSEARCH_CI | 1<<QSEARCH_JOKER_CS | 1<<QSEARCH_JOKER_CI,
	&simple_constructor,
	&simple_destructor,
	&simple_add_pattern,
	&simple_compile,
	&simple_search,
	&simple_alloc_io_state,
	&simple_free_state,
};



simple_algo_t *QSEARCH_ALGO_LIST_SYM[] = {
	&algo_simple,
	NULL
};

int QSEARCH_API_VERSION_SYM = QSEARCH_API_VERSION;
