/*
 * qsearch.h -- qsearch API header file.
 *              see http://www.cartel-securite.net/pbiondi/libqsearch.html
 *              for more informations
 *
 * Copyright (C) 2002  Philippe Biondi <biondi@cartel-securite.fr>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */


#ifndef QSEARCH_H
#define QSEARCH_H

#ifdef __KERNEL__
#include <linux/slab.h>
#else /* __KERNEL__ */
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#endif /* __KERNEL__ */

#define QSEARCH_STR2(x) #x
#define QSEARCH_STR(X) QSEARCH_STR2(X)

#ifdef DEBUG
#ifdef __KERNEL__
#define DBG(args...) printk("##"__FUNCTION__".l" QSEARCH_STR(__LINE__) ": " ##args);
#else  /* __KERNEL__ */
#define DBG(args...) printf("##"__FUNCTION__".l" QSEARCH_STR(__LINE__) ": " ##args);
#endif /* __KERNEL__ */
#else
#define DBG(args...)
#endif


/* Version */
#define QSEARCH_API_VERSION 0x091

/* Symbol used for algorithm lists */
#define QSEARCH_ALGO_LIST_SYM algo_list
#define QSEARCH_API_VERSION_SYM qsearch_api_version

/* Pattern types */
#define QSEARCH_CS               0 /* normal case sensitive pattern */
#define QSEARCH_CI               1 /* case insensitive pattern */
#define QSEARCH_JOKER_CS         2 /* pattern using '*' for any char */
#define QSEARCH_JOKER_CI         3 /* pattern using '*' for any char */
#define QSEARCH_REGEXP           4 /* regular expression */

/* Error types */
#define QENOMEM          1  /* out of memory */
#define QEBADTYPE        2  /* bad pattern type */
#define QENOTIMPLEMENTED 3  /* function not implemented */
#define QEINTERNAL       4  /* internal error */
#define QESTATES         5  /* the use of states is not supported by the algo */
	    

/* Definitions of objects algo, obj and options */
#define QSEARCH_ALGO(algo)                                                              \
        const char *name;                                                               \
        const char *info;        						        \
	unsigned int supported_types;                                                   \
        algo##_obj_t *(*constructor)(algo##_algo_t *algo, algo##_options_t *options);   \
        int (*destructor)(algo##_obj_t *self);           			        \
        int (*add_pattern)(algo##_obj_t *self, unsigned char *pattern,                  \
			   int len, void *ptrn_data, int type);                         \
        int (*compile)(algo##_obj_t *self);			  		        \
        int (*search)(algo##_obj_t *self, 					        \
                      algo##_state_t *state_in, algo##_state_t **state_out,             \
                      algo##_state_t *state_io,                                         \
                      qsearch_callback_t cb, void *cb_data,                             \
                      void *buffer, size_t len);                                        \
        int (*alloc_io_state)(algo##_obj_t *self, algo##_state_t **state);              \
	int (*free_state)(algo##_obj_t *self, algo##_state_t **state);                   

#define QSEARCH_OBJ(algo)           \
            algo##_algo_t *statics; \
            int state;

#define QSEARCH_OPTIONS(algo) \
	char joker;



/* The callback function gets cb_data from the qsearch_search() caller.
 * ptrn_data is the one associated with the pattern that matched.
 * match_offset is the offset of the *last* byte of the pattern in the buffer
 */
typedef int (*qsearch_callback_t)(void *cb_data, void *ptrn_data, size_t match_offset);
typedef struct {} qsearch_state_t;
typedef struct qsearch_obj_s qsearch_obj_t;
typedef struct qsearch_algo_s qsearch_algo_t;
typedef struct qsearch_options_s qsearch_options_t;

struct qsearch_algo_s {
        QSEARCH_ALGO(qsearch)
};

struct qsearch_obj_s {
	QSEARCH_OBJ(qsearch)
};

struct qsearch_options_s {
	QSEARCH_OPTIONS(qsearch)
};

#ifdef __KERNEL__
typedef struct qsearch_plugin_s {
	char *name;
	void (*add_use_count)(int val);
	qsearch_algo_t **algo_list;
} qsearch_plugin_t;
#else /* __KERNEL__ */
typedef struct qsearch_plugin_s {
	void *lib_handle;
	qsearch_algo_t **algo_list;
} qsearch_plugin_t;
#endif /* __KERNEL__ */


/* Get the list of available plugins in QSEARCH_PLUGIN_DIR */
char **qsearch_get_plugin_list(void);

/* Free a plugin list allocated by qsearch_get_plugin_list() */
void qsearch_free_plugin_list(char **list);

/* Load a plugin */
qsearch_plugin_t *qsearch_load_plugin(const char *name);

/* Unload a previously loaded plugin */
int qsearch_unload_plugin(qsearch_plugin_t *plugin);

/* Get an algo from a plugin */
qsearch_algo_t *qsearch_get_algo(qsearch_plugin_t *plugin, const char *name);

/* Get the list of algorithms provided by a plugin */
qsearch_algo_t **qsearch_get_algo_list(qsearch_plugin_t *plugin);

/* Instantiate a search object from a given algorithm */
qsearch_obj_t *qsearch_constructor(qsearch_algo_t *algo, qsearch_options_t *options);

/* Destruct a search object */
int qsearch_destructor(qsearch_obj_t *self);

/* Add a search pattern to a search object, and associate pattern data to it */
int qsearch_add_pattern(qsearch_obj_t *self, unsigned char *pattern, int len, void *ptrn_data, int type);

/* Once each search pattern has been added to the object, compile it */
int qsearch_compile(qsearch_obj_t *self);

/* Once the object is compiled, search the buffer, using state_in as start state.
 * The very first one state is NULL. Each time a match occurs, the callback function
 * is called with cb_data. If the callback return is > 0 the search ends and the 
 * result is returned. Else the function returns 0. The state_out variable is allocated and
 * initialized and can be used as state_in to carry on the search.
 * Alternately, you can alloc a state_io with qsearch_alloc_io_state(), that will be 
 * used as input state and output state. You have to provide NULL to state_in and state_out
 * and use the allocated buffer for state_io.
 * A result < 0 means an error (malloc,...)
 */
int qsearch_search(qsearch_obj_t *self,
                   qsearch_state_t *state_in, qsearch_state_t **state_out,
                   qsearch_state_t *state_io,
                   qsearch_callback_t cb, void *cb_data,
                   void *buffer, size_t len);

/* Allocate a state that is big enough to contain any states and that can be used
 * as a state_io in qsearch_search(), i.e. it will be used as both state_in and
 * state_out (if the latters are given as NULL)
 */
int qsearch_alloc_io_state(qsearch_obj_t *self, qsearch_state_t **state);

/* Free a state that has been allocated by qsearch_search() or 
 * qsearch_alloc_io_state()
 */
int qsearch_free_state(qsearch_obj_t *self, qsearch_state_t **state);

#ifdef __KERNEL__
/* Register and deregister a plugin */
int qsearch_register(qsearch_plugin_t *plugin);
int qsearch_deregister(qsearch_plugin_t *plugin);
#endif


#endif /* QSEARCH_H */
