/*
 * qsearch.c -- qsearch API implementation
 *              see http://www.cartel-securite.net/pbiondi/libqsearch.html
 *              for more informations
 *
 * Copyright (C) 2002  Philippe Biondi <biondi@cartel-securite.fr>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>


#include "ltdl.h"
#include "qsearch.h"

#define STR2(x) #x
#define STR(x) STR2(x)

void _init(void);

void _init(void)
{
        if ( lt_dlinit() < 0 ) {
        	DBG("Error initializing libltdl: %s.\n", lt_dlerror());
        }
}


char **qsearch_get_plugin_list(void)
{
	DIR *dir;
	struct dirent *d;
	char *name,*dot;
	char **list;
	int nb;

	dir = opendir(QSEARCH_PLUGIN_DIR);
	if (!dir) {
		DBG("opendir(%s): error %i\n", QSEARCH_PLUGIN_DIR, errno);
		return NULL;
	}

        list = malloc(sizeof(char *));
        
	*list = NULL;
	nb = 0;
	
	while ((d = readdir(dir))) {
                
		dot = strrchr(d->d_name, '.');
		if ( ! dot || strcmp(dot + 1, "la") )
			continue;
                
		*dot = 0;
		name = strdup(d->d_name);
                
		list = realloc(list, (++nb + 1) * sizeof(char *));
		list[nb - 1] = name;
		list[nb] = NULL;
	}
        
	closedir(dir);

        return list;
}

void qsearch_free_plugin_list(char **list)
{
	char **p;

	if (list) {
		p=list;
		while (*p) {
			free(*p);
			p++;
		}
		free(list);
	}
}

qsearch_plugin_t *qsearch_load_plugin(const char *name)
{
        void *handle;
	int *api_ver;
        char filename[1024];
        qsearch_algo_t **list;
        qsearch_plugin_t *plugin;

	/* 
	 * Create the lib name
	 */
        if ( ! strchr(name, '/') ) {
                snprintf(filename, sizeof(filename), "%s/%s", QSEARCH_PLUGIN_DIR , name);
                name = filename;
        }
                
	/* 
	 * Load the lib
	 */
        DBG("loading lib [%s]\n", name);
        handle = lt_dlopenext(name);
	if ( ! handle ) {
		DBG("Error loading [%s]: %s\n", namename, lt_dlerror())
		goto err_load;
	}
        
       	/* 
	 * Retrieve the qsearch_api_version symbol and check the API version
	 */	
	api_ver = lt_dlsym(handle, STR(QSEARCH_API_VERSION_SYM));
	if ( ! api_ver ) {
		DBG("Error loading symbol '" STR(QSEARCH_API_VERSION_SYM) "': %s\n", lt_dlerror());
		goto err_closelib;
	}
	if (*api_ver != QSEARCH_API_VERSION) {
		DBG("Error loading plugin '%s': API version mismatch (%s is %#x and libqsearch is %#x)\n",
		    name, name, *api_ver, QSEARCH_API_VERSION);
		goto err_closelib;
	}
		

       	/* 
	 * Retrieve the algo_list symbol
	 */	
	list = lt_dlsym(handle, STR(QSEARCH_ALGO_LIST_SYM));
	if ( ! list ) {
		DBG("Error loading symbol '" STR(QSEARCH_ALGO_LIST_SYM) "': %s\n", lt_dlerror());
		goto err_closelib;
	}

	/* 
	 * Create the plugin object
	 */
        plugin=malloc(sizeof(qsearch_plugin_t));
	if (!plugin) {
		DBG("Error: malloc failed for plugin.\n");
		goto err_closelib;
	}
	plugin->algo_list = list;
	plugin->lib_handle = handle;
	
        return plugin;

err_closelib:	
	lt_dlclose(handle);
err_load:
	return NULL;
}

qsearch_algo_t *qsearch_get_algo(qsearch_plugin_t *plugin, const char *name)
{
        qsearch_algo_t **list;

	/*
	 * Find the algo by its name
	 */	
	list=plugin->algo_list;
	while (*list) {
		if (! strcmp( (*list)->name, name)) 
			break;
		list++;
	}

	return *list;
}

qsearch_algo_t **qsearch_get_algo_list(qsearch_plugin_t *plugin)
{
	return plugin->algo_list;
}

int qsearch_unload_plugin(qsearch_plugin_t *plugin)
{
	void *handle;
	
	handle = plugin->lib_handle;
	free(plugin);

	return lt_dlclose(handle);
	
}

/*
 * Wrapper functions 
 */

qsearch_obj_t *qsearch_constructor(qsearch_algo_t *algo, qsearch_options_t *options) 
{
	return (*algo->constructor)(algo, options);
}

int qsearch_destructor(qsearch_obj_t *self)
{
	return (*self->statics->destructor)(self);
}

int qsearch_add_pattern(qsearch_obj_t *self, unsigned char *pattern, int len, void *ptrn_data, int type)
{
	if ((1<<type) & self->statics->supported_types)
		return (*self->statics->add_pattern)(self,pattern,len,ptrn_data,type);
	return -QEBADTYPE;
}

int qsearch_compile(qsearch_obj_t *self)
{
	return (*self->statics->compile)(self);
}

int qsearch_search(qsearch_obj_t *self,
		   qsearch_state_t *state_in, qsearch_state_t **state_out,
		   qsearch_state_t *state_io, 
		   qsearch_callback_t cb, void *cb_data,
		   void *buffer, size_t len)                                  
{
	return (*self->statics->search)(self, state_in, state_out, state_io, 
					cb, cb_data, buffer, len);
}

int qsearch_alloc_io_state(qsearch_obj_t *self, qsearch_state_t **state)
{	
  	return (*self->statics->alloc_io_state)(self, state);
}

int qsearch_free_state(qsearch_obj_t *self, qsearch_state_t **state)
{	
  	return (*self->statics->free_state)(self, state);
}
