/*
 * testsuite.c -- qsearch testsuite
 *                see http://www.cartel-securite.net/pbiondi/libqsearch.html
 *                for more informations
 *
 * Copyright (C) 2002  Philippe Biondi <biondi@cartel-securite.fr>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <dlfcn.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include "qsearch.h"
#include "testsuite.h"

#ifndef CONFDIR
#define CONFDIR ""
#endif

static int cb(void *cb_data, void *ptrn_data, size_t match_offset)
{
	(*((char *)ptrn_data+match_offset-1))++;
	return 0;
}


test_t **test_list;

#define SKIPPED "   [S] "
#define ERROR   "   [E] "
#define FAILED  "   [F] "

#define ROK   0
#define RSKIP 1 
#define RFAILED 2
#define RERROR 3

char normalcol[] = "\033[0m";
int curcol = 0;
int nbcol = 2;
char *color[] = {
	"\033[02;34m",
	"\033[02;31m",
};

static int do_test(test_t *test, qsearch_algo_t *algo, int frag)
{
	ptrn_t **pat;
        qsearch_obj_t *obj;
	int patnb,len;
	char **result, **presult, **good;
	int ret, failed, err=ROK;
	
	printf("%s",color[curcol++%nbcol]);

	printf("[   ?   ] [FRAG=%+02i] %s (%s)\r",frag, test->name, test->info); fflush(stdout);

	obj = qsearch_constructor(algo, NULL);

	len = strlen(test->sample);
	patnb=0;
	for (pat = test->patterns; *pat; pat++) { patnb++; }

	presult = result = malloc(patnb*sizeof(char *));
	
	/* Adding the patterns */
	for (pat = test->patterns; *pat; pat++, presult++) {
		(*presult)=malloc(len+1);
		memset(*presult, '0', len);
		*((*presult)+len)='\0';
		ret=qsearch_add_pattern(obj, (*pat)->pattern, 
		  		        strlen((*pat)->pattern),
				        *presult, (*pat)->type);
		if (ret == -QEBADTYPE) {
			printf("[Skipped]\n" SKIPPED "The algorithm does not support patterns of type %i\n",(*pat)->type);
			err=RSKIP;
			goto finish_test;
		}
		else if (ret) {
			printf("[ ERROR ]\n" ERROR "An error (%i) occured while adding the pattern [%s]\n",ret,(*pat)->pattern);
			err=RERROR;
			goto finish_test;
		}
	}

	/* Compile the search object */
	ret=qsearch_compile(obj);
	if (ret) {
		printf("[ ERROR ]\n" ERROR "An error (%i) occured while compiling the seach object\n",ret);
		err=RERROR;
		goto finish_test;
	}

	/* Searching */
	if (frag > 0) {  /* Tests with state_in and state_out */
		qsearch_state_t *state_in, *state_out;
	        int pos;
		
		state_in=NULL;
		state_out=NULL;
		pos=0;
		ret=0;
		while (!ret && (pos < len)) {
			ret=qsearch_search(obj, state_in, &state_out, NULL, &cb, NULL, 
					   test->sample+pos, pos+frag > len ? len%frag : frag);
			qsearch_free_state(obj, &state_in);
			state_in=state_out;
			pos += frag;
		}
		qsearch_free_state(obj, &state_in);
	}
	else if (frag < 0) {  /* Tests with state_io */
		qsearch_state_t *state_io;
	        int pos;
		
		frag = -frag;
		pos = 0;
		ret = 0;
		if ((ret = qsearch_alloc_io_state(obj, &state_io))) {
			switch (ret) {
			case -QENOTIMPLEMENTED:
				printf("[Skipped]\n" SKIPPED "The algorithm does not support state_io\n");
				err = RSKIP;
				break;
			default:
				printf("[ ERROR ]\n" ERROR "An error (%i) occured allocating state_io\n",ret);
				err = RERROR;
			}
			goto finish_test;
		}

		while (!ret && (pos < len)) {
			ret=qsearch_search(obj, NULL, NULL, state_io, &cb, NULL, 
					   test->sample+pos, pos+frag > len ? len%frag : frag);
			pos += frag;
		}
		qsearch_free_state(obj, &state_io);
	}
	else {
		ret=qsearch_search(obj, NULL, NULL, NULL, &cb, NULL, test->sample, len);
	}
	if (ret) {
		printf("[ ERROR ]\n" ERROR "An error (%i) occured while searching\n",ret);
		err = RERROR;
		goto finish_test;
	}


	/* Checking the results */
	failed=0;
	presult=result;
	pat=test->patterns;
       	good=test->results;
       	for ( ; *good; presult++, pat++, good++) {
		if (!strcmp(*good, *presult))
			continue;
		if (!failed) printf("[FAILED!]\n");
		failed++;
		printf(FAILED "failed for [%s] searching [%s]\n" FAILED "Found      |%s|\n" FAILED "instead of |%s|\n",
		      test->sample, (*pat)->pattern, *presult, *good);
		
	}
	if (failed) {
		err=RFAILED;
	}
	else {
		err=ROK;
	}
	
	
finish_test:
	ret=qsearch_destructor(obj);
	if (ret) {
		printf("[ERROR]\n" ERROR "An error (%i) occured while destroying the object\n",ret);
		err=RERROR;
	}
	for (presult=result; patnb > 0; patnb--, presult++) {
		free(*presult);
	}
	free(result);
	if (err == ROK) 
		printf("[  OK!  ]\n");
	printf("%s", normalcol);
	return err;
}





static void testsuite(char *testname, qsearch_algo_t *algo, int frag)
{
	test_t **test;
	int i,len;
	int stats[4]={0,0,0,0};

	printf("** Test for algo [%s]\n",algo->name);
	for (test = test_list; *test; test++) {
		if ((testname) && (strcmp((*test)->name,testname)))
				continue;
		len=strlen((*test)->sample);
		stats[do_test(*test, algo, 0)]++;
		if (!frag) continue;
		for (i=1; i <= len; i++) {
			stats[do_test(*test, algo, i)]++;
			stats[do_test(*test, algo, -i)]++;
		}
	}
	printf("==================\nSummary: OK=%i  SKIPPED=%i  FAILED=%i  ERROR=%i\n",
	       stats[ROK], stats[RSKIP], stats[RFAILED], stats[RERROR]);

}	
		
static void usage()
{
    	fprintf(stderr,"Error: Usage: testsuite -p <plugin> [-a <algo>] [-t <test>] [-f <conffile>]\n\t[-D]");
	exit(1);
}

int main(int argc, char **argv) 
{
	qsearch_plugin_t *plugin;
	qsearch_algo_t *algo, **list;


	int c;
	char *algoname=NULL, *pluginname=NULL, *testname=NULL;
	char *conffile=CONFDIR "testsuite.conf";
	int frag=1;


	/* Parsing arguments line */
	while ((c = getopt(argc,argv, "Dp:a:f:t:")) != -1) {
		switch (c) {
			case 'p':
				pluginname=optarg;
				break;
			case 'a':
				algoname=optarg;
				break;
			case 'f':
				conffile=optarg;
				break;
			case 't':
				testname=optarg;
				break;
			case 'D':
				frag=0;
				break;
			default:
				usage();
		}
	}

	/* Read config file */
	test_list=readconf(conffile);
	
	/* Checking and parsing arguments */	
	if (!pluginname) usage();
	

	/* Loading the plugin */
	printf("Loading plugin '%s'.\n",pluginname);
	plugin=qsearch_load_plugin(pluginname);
	if (!plugin) {
	    fprintf(stderr,"Error: plugin '%s' not found\n", pluginname);
	    exit(3);
	}

	/* Displaying the availables algorithms from the plugin */
	list=qsearch_get_algo_list(plugin);
	if (! *list) {
		fprintf(stderr,"Error plugin '%s' does not provide any algorithms\n",pluginname);
		exit(4);
	}
	printf("Available algorithms in [%s]:\n", pluginname);
	for ( ; *list; list++) {
		printf(" + %s [%0x] (%s)\n",(*list)->name, (*list)->supported_types,(*list)->info);
	}

	/* Getting an algorithm from the plugin and testing it */
	if (algoname) {
		algo=qsearch_get_algo(plugin, algoname);
		if (! algo) {
        	        fprintf(stderr,"Error algorithm '%s' not found\n",algoname);
                	exit(4);
		}
		testsuite(testname,algo,frag);
	}
	else {
		list=qsearch_get_algo_list(plugin);
		while (*list) {
			testsuite(testname,*list,frag);
			list++;
		}
	}

	/* Leaving */
	qsearch_unload_plugin(plugin);
	printf("Testsuite done.\n");
	return 0;
}

